import contextlib
import subprocess
import sys
import sysconfig
import textwrap
from typing import TextIO

from pydust import config

PYVER_MINOR = ".".join(str(v) for v in sys.version_info[:2])
PYVER_HEX = f"{sys.hexversion:#010x}"
PYINC = sysconfig.get_path("include")
PYLIB = sysconfig.get_config_var("LIBDIR")


def zig_build(argv: list[str]):
    conf = config.load()
    generate_build_zig(conf.build_zig)
    subprocess.run(
        [sys.executable, "-m", "ziglang", "build", "--build-file", conf.build_zig] + argv,
        check=True,
    )


def generate_build_zig(build_zig_file):
    """Generate the build.zig file for the current pyproject.toml.

    Initially we were calling `zig build-lib` directly, and this worked fine except it meant we
    would need to roll our own caching and figure out how to convince ZLS to pick up our dependencies.

    It's therefore easier, at least for now, to generate a build.zig in the project root and add it
    to the .gitignore. This means ZLS works as expected, we can leverage zig build caching, and the user
    can inspect the generated file to assist with debugging.
    """
    conf = config.load()

    with open(build_zig_file, "w+") as f:
        b = Writer(f)

        b.writeln('const std = @import("std");')
        b.writeln()

        # We choose to ship Pydust source code inside our PyPi package. This means the that once the PyPi package has
        # been downloaded there are no further requests out to the internet.
        #
        # As a result of this though, we need some way to locate the source code. We can't use a reference to the
        # currently running pydust, since that could be inside a temporary build env which won't exist later when the
        # user is developing locally. Therefore, we defer resolving the path until the build.zig is invoked by shelling
        # out to Python.
        #
        # We also want to avoid creating a pydust module (and instead prefer anonymous modules) so that we can populate
        # a separate pyconf options object for each Python extension module.
        b.write(
            """
            fn getPydustRootPath(allocator: std.mem.Allocator) ![]const u8 {
                const includeResult = try std.process.Child.exec(.{
                    .allocator = allocator,
                    .argv = &.{
                        "python",
                        "-c",
                        \\\\import os
                        \\\\import pydust
                        \\\\print(os.path.join(os.path.dirname(pydust.__file__), 'src', 'pydust.zig'), end='')
                        \\\\
                    },
                });
                allocator.free(includeResult.stderr);
                return includeResult.stdout;
            }
            """
        )
        b.writeln()

        with b.block("pub fn build(b: *std.Build) void"):
            b.write(
                """
                const target = b.standardTargetOptions(.{});
                const optimize = b.standardOptimizeOption(.{});

                const test_step = b.step("test", "Run library tests");
                const test_build_step = b.step("test-build", "Build test runners");

                const pydust = getPydustRootPath(b.allocator) catch @panic("Failed to locate Pydust source code");
                """
            )

            for ext_module in conf.ext_modules:
                # TODO(ngates): fix the out filename for non-limited modules
                assert ext_module.limited_api, "Only limited_api is supported for now"

                # For each module, we generate some options, a library, as well as a test runner.
                with b.block():
                    b.write(
                        f"""
                        // For each Python ext_module, generate a shared library and test runner.
                        const pyconf = b.addOptions();
                        pyconf.addOption([:0]const u8, "module_name", "{ext_module.name}");
                        pyconf.addOption(bool, "limited_api", {str(ext_module.limited_api).lower()});
                        pyconf.addOption([:0]const u8, "hexversion", "{PYVER_HEX}");

                        const lib{ext_module.libname} = b.addSharedLibrary(.{{
                            .name = "{ext_module.libname}",
                            .root_source_file = .{{ .path = "{ext_module.root}" }},
                            .main_pkg_path = .{{ .path = "{conf.root}" }},
                            .target = target,
                            .optimize = optimize,
                        }});
                        configurePythonInclude(pydust, lib{ext_module.libname}, pyconf);

                        // Install the shared library within the source tree
                        const install{ext_module.libname} = b.addInstallFileWithDir(
                            lib{ext_module.libname}.getEmittedBin(),
                            .{{ .custom = ".." }},  // Relative to project root: zig-out/../
                            "{ext_module.install_path}",
                        );
                        b.getInstallStep().dependOn(&install{ext_module.libname}.step);

                        const test{ext_module.libname} = b.addTest(.{{
                            .root_source_file = .{{ .path = "{ext_module.root}" }},
                            .main_pkg_path = .{{ .path = "{conf.root}" }},
                            .target = target,
                            .optimize = optimize,
                        }});
                        configurePythonRuntime(pydust, test{ext_module.libname}, pyconf);

                        // Install the test binary
                        const installtest{ext_module.libname} = b.addInstallBinFile(
                            test{ext_module.libname}.getEmittedBin(),
                            "{ext_module.libname}.test.bin",
                        );
                        test_build_step.dependOn(&installtest{ext_module.libname}.step);

                        // Run the tests as part of zig build test.
                        const run_test{ext_module.libname} = b.addRunArtifact(test{ext_module.libname});
                        test_step.dependOn(&run_test{ext_module.libname}.step);
                        """
                    )

            b.write(
                f"""
                // Option for emitting test binary based on the given root source. This can be helpful for debugging.
                const debugRoot = b.option(
                    []const u8,
                    "debug-root",
                    "The root path of a file emitted as a binary for use with the debugger",
                );
                if (debugRoot) |root| {{
                    const pyconf = b.addOptions();
                    pyconf.addOption([:0]const u8, "module_name", "debug");
                    pyconf.addOption(bool, "limited_api", false);
                    pyconf.addOption([:0]const u8, "hexversion", "{PYVER_HEX}");

                    const testdebug = b.addTest(.{{
                        .root_source_file = .{{ .path = root }},
                        .main_pkg_path = .{{ .path = "{conf.root}" }},
                        .target = target,
                        .optimize = optimize,
                    }});
                    configurePythonRuntime(pydust, testdebug, pyconf);

                    const debugBin = b.addInstallBinFile(testdebug.getEmittedBin(), "debug.bin");
                    b.getInstallStep().dependOn(&debugBin.step);
                }}
                """
            )

        b.write(
            f"""
            fn configurePythonInclude(
                pydust: []const u8, compile: *std.Build.CompileStep, pyconf: *std.Build.Step.Options,
            ) void {{
                compile.addAnonymousModule("pydust", .{{
                    .source_file = .{{ .path = pydust }},
                    .dependencies = &.{{.{{ .name = "pyconf", .module = pyconf.createModule() }}}},
                }});
                compile.addIncludePath(.{{ .path = "{PYINC}" }});
                compile.linkLibC();
                compile.linker_allow_shlib_undefined = true;
            }}

            fn configurePythonRuntime(
                pydust: []const u8, compile: *std.Build.CompileStep, pyconf: *std.Build.Step.Options
            ) void {{
                configurePythonInclude(pydust, compile, pyconf);
                compile.linkSystemLibrary("python{PYVER_MINOR}");
                compile.addLibraryPath(.{{ .path =  "{PYLIB}" }});
            }}
            """
        )


class Writer:
    def __init__(self, fileobj: TextIO) -> None:
        self.f = fileobj
        self._indent = 0

    @contextlib.contextmanager
    def indent(self):
        self._indent += 4
        yield
        self._indent -= 4

    @contextlib.contextmanager
    def block(self, text: str = ""):
        self.write(text)
        self.writeln(" {")
        with self.indent():
            yield
        self.writeln()
        self.writeln("}")
        self.writeln()

    def write(self, text: str):
        if "\n" in text:
            text = textwrap.dedent(text).strip() + "\n\n"
        self.f.write(textwrap.indent(text, self._indent * " "))

    def writeln(self, text: str = ""):
        self.write(text)
        self.f.write("\n")


if __name__ == "__main__":
    generate_build_zig("test.build.zig")



import enum
import io
import pathlib
import struct
import subprocess
import tempfile

import pytest
from pydantic import BaseModel

from pydust import buildzig, config

pydust_conf = config.load()


def pytest_collection(session):
    """Invoke a `zig build test-build` once for all of our modules at the start of the session.

    This is our own build target that generates the test runners but doesn't invoke them.

    This is fine due to zig build's caching. We then run test collection for each of the Pydust
    entrypoints in the pyproject.toml.
    """
    buildzig.zig_build(["test-build"])


def pytest_collect_file(file_path, path, parent):
    """Grab any Zig roots for PyTest collection."""
    if file_path.suffix == ".zig":
        for ext_module in pydust_conf.ext_modules:
            if pathlib.Path(ext_module.root).absolute() == file_path.absolute():
                return ZigFile.from_parent(parent, path=file_path)


class ZigFile(pytest.File):
    def collect(self):
        """Collect all the tests that exist within this Zig root.

        First compile using 'zig test'
        Then spin up the test server and query it for the test metadata.
        """
        ext_module = [e for e in pydust_conf.ext_modules if pathlib.Path(e.root).absolute() == self.path][0]

        # Then query the test metadata
        proc = subprocess.Popen([ext_module.test_bin, "--listen=-"], stdin=subprocess.PIPE, stdout=subprocess.PIPE)
        try:
            # Zig first sends us its version.
            h = TestProtocol.Header.unpack(proc.stdout)
            assert h.tag == TestProtocol.ResponseTag.zig_version.value
            _zig_version = proc.stdout.read(h.bytes_len).decode("utf-8")

            # Then we can request test metadata
            proc.stdin.write(TestProtocol.Header(tag=TestProtocol.RequestTag.query_test_metadata.value).pack())
            proc.stdin.flush()

            h = TestProtocol.Header.unpack(proc.stdout)
            assert h.tag == TestProtocol.ResponseTag.test_metadata.value
            test_metas = self._read_test_metadata(proc.stdout.read(h.bytes_len))
        finally:
            proc.kill()
            proc.wait()

        for test in test_metas:
            # TODO(ngates): we could override the path here if test_metadata provided source provenance.
            yield ZigItem.from_parent(self, name=test["name"], ext_module=ext_module, test_meta=test)

    @staticmethod
    def _read_test_metadata(buffer):
        (string_bytes_len, tests_len) = struct.unpack("<II", buffer[:8])
        buffer = buffer[8:]

        # Extract byte offsets into the test metadata where each of the values occurs.
        # End offsets are null-terminated bytes.
        # Wrap the buffer so we can consume from it nicely in a loop
        fileobj = io.BytesIO(buffer)
        names = [struct.unpack("<I", fileobj.read(4))[0] for _ in range(tests_len)]
        async_frame_sizes = [struct.unpack("<I", fileobj.read(4))[0] for _ in range(tests_len)]
        expected_panic_msgs = [struct.unpack("<I", fileobj.read(4))[0] for _ in range(tests_len)]

        # We use the original buffer to extract string data since it's easier to find the next null terminator
        data = buffer[-string_bytes_len:]
        tests = []
        for i, (name, afs, ep) in enumerate(zip(names, async_frame_sizes, expected_panic_msgs)):
            test_name = data[name : data.index(b"\0", name)].decode("utf-8")
            if test_name.startswith("test."):
                test_name = test_name[len("test.") :]
            tests.append(
                {
                    "idx": i,
                    "name": test_name,
                    "async_frame_size": data[afs : data.index(b"\0", afs)].decode("utf-8") if afs else None,
                    "expected_panics": data[ep : data.index(b"\0", ep)].decode("utf-8") if ep else None,
                }
            )

        return tests


class ZigItem(pytest.Item):
    def __init__(self, *, ext_module, test_meta, **kwargs):
        super().__init__(**kwargs)
        self.ext_module = ext_module
        self.test_meta = test_meta

        self._stderr = None

    def runtest(self):
        stderr = tempfile.NamedTemporaryFile()
        proc = subprocess.Popen(
            [self.ext_module.test_bin, "--listen=-"],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=stderr,
        )
        try:
            # Zig first sends us its version.
            h = TestProtocol.Header.unpack(proc.stdout)
            assert h.tag == TestProtocol.ResponseTag.zig_version.value
            _zig_version = proc.stdout.read(h.bytes_len).decode("utf-8")

            # Then we can request the test to run
            proc.stdin.write(TestProtocol.Header(tag=TestProtocol.RequestTag.run_test.value, bytes_len=4).pack())
            proc.stdin.write(struct.pack("<I", self.test_meta["idx"]))
            proc.stdin.flush()

            h = TestProtocol.Header.unpack(proc.stdout)
            assert h.tag == TestProtocol.ResponseTag.test_results.value

            _test_idx, flags = struct.unpack("<II", proc.stdout.read(8))
            fail = bool(flags & 0x01)
            skip = bool(flags & 0x02)
            leak = bool(flags & 0x04)
            # TODO(ngates): log_error_count: u29 isn't currently passed back but should be

            with open(stderr.name) as f:
                self.add_report_section("call", "stderr", f.read())
        except Exception as e:
            raise Exception("Zig test crashed. Exited " + str(proc.wait())) from e
        finally:
            proc.kill()
            proc.wait()
            stderr.close()

        if skip:
            self.add_marker(pytest.mark.skip)

        if leak:
            self.add_report_section("call", "memory leaks", f"Zig detected a memory leak in '{self.nodeid}'")

        if fail or leak:
            raise Exception("Failure in Zig test")

    def repr_failure(self, excinfo):
        """Called when self.runtest() raises an exception."""
        return str(excinfo)

    def reportinfo(self):
        return self.path, 0, self.test_meta["name"]


class TestProtocol:
    """Utilities for reading and writing messages to the Zig test runner server."""

    class RequestTag(enum.Enum):
        # Tells the compiler to shut down cleanly.
        # No body.
        exit = 0
        # Tells the compiler to detect changes in source files and update the
        # affected output compilation artifacts.
        # If one of the compilation artifacts is an executable that is
        # running as a child process, the compiler will wait for it to exit
        # before performing the update.
        # No body.
        update = 1
        # Tells the compiler to execute the executable as a child process.
        # No body.
        run = 2
        # Tells the compiler to detect changes in source files and update the
        # affected output compilation artifacts.
        # If one of the compilation artifacts is an executable that is
        # running as a child process, the compiler will perform a hot code
        # swap.
        # No body.
        hot_update = 3
        # Ask the test runner for metadata about all the unit tests that can
        # be run. Server will respond with a `test_metadata` message.
        # No body.
        query_test_metadata = 4
        # Ask the test runner to run a particular test.
        # The message body is a u32 test index.
        run_test = 5

    class ResponseTag(enum.Enum):
        # Body is a UTF-8 string.
        zig_version = 0
        # Body is an ErrorBundle.
        error_bundle = 1
        # Body is a UTF-8 string.
        progress = 2
        # Body is a EmitBinPath.
        emit_bin_path = 3
        # Body is a TestMetadata
        test_metadata = 4
        # Body is a TestResults
        test_results = 5

    class Header(BaseModel):
        tag: int
        bytes_len: int = 0

        def pack(self):
            return struct.pack("<II", self.tag, self.bytes_len)

        @classmethod
        def unpack(cls, buffer):
            (tag, bytes_len) = struct.unpack("<II", buffer.read(8))
            return cls(tag=tag, bytes_len=bytes_len)



import argparse

from pydust import buildzig

parser = argparse.ArgumentParser()
sub = parser.add_subparsers(dest="command", required=True)

debug = sub.add_parser("debug", help="Compile a Zig file with debug symbols. Useful for running from an IDE.")
debug.add_argument("entrypoint")


def main():
    args = parser.parse_args()

    if args.command == "debug":
        debug(args)


def debug(args):
    """Given an entrypoint file, compile it for test debugging. Placing it in a well-known location."""
    entrypoint = args.entrypoint
    buildzig.zig_build(["install", f"-Ddebug-root={entrypoint}"])


if __name__ == "__main__":
    main()



import functools
import os

import tomllib
from pydantic import BaseModel, Field


class ExtModule(BaseModel):
    """Config for a single Zig extension module."""

    name: str
    root: str
    limited_api: bool = True

    @property
    def libname(self):
        return self.name.rsplit(".", maxsplit=1)[-1]

    @property
    def install_path(self):
        # FIXME(ngates): for non-limited API
        assert self.limited_api, "Only limited API modules are supported right now"
        return os.path.join(*self.name.split(".")) + ".abi3.so"

    @property
    def test_bin(self):
        return os.path.join("zig-out", "bin", self.libname + ".test.bin")


class ToolPydust(BaseModel):
    """Model for tool.pydust section of a pyproject.toml."""

    root: str = "src/"

    build_zig: str = "build.zig"

    # We rename pluralized config sections so the pyproject.toml reads better.
    ext_modules: list[ExtModule] = Field(alias="ext_module", default_factory=list)


@functools.cache
def load() -> ToolPydust:
    with open("pyproject.toml", "rb") as f:
        pyproject = tomllib.load(f)
    return ToolPydust(**pyproject["tool"].get("pydust", {}))
